#ifndef CORE_INTEGER_H
#define CORE_INTEGER_H

#include "System.h"

namespace rdecore
{
/// Integer number related functions.
namespace Integer
{
/// Returns the lowest 8-bits of number.
inline uint8_t LoByte(uint32_t x)
{
    return uint8_t(x & 0xFF);
}    

/// Returns higher 8 bits of given number.
inline uint8_t HiByte(uint32_t x)
{
    return uint8_t((x & 0xFF00) >> 8);
}

/// Returns the lowest 16 bits of given number.
inline uint16_t LoWord(uint32_t x)
{
    return uint16_t(x & 0xFFFF);
}

/// Returns higher 16 bits of given number.
inline uint16_t HiWord(uint32_t x)
{
    return uint16_t((x & 0xFFFF0000) >> 16);
}

//-----------------------------------------------------------------------------    
/**
 * Returns bottom bit position of given mask (zero-based index of first 
 * non zero bit).
 * @par Example:
 *      - \c BottomBit(0xFF)    returns 0
 *      - \c BottomBit(0xFF00)  return 8
 */
inline uint32_t BottomBit(uint32_t x)
{
    if (x == 0)
        return 0;
    
    uint32_t bottomBit(0);
    while (!(x & 0x1))
    {
        x >>= 1;
        ++bottomBit;
    }
    return bottomBit;
}

//-----------------------------------------------------------------------------    
/**
 * Returns top bit position of given mask (zero-based index of last non-zero
 * bit).
 * @par Example:
 *      - \c TopBit(0xFF)   returns 8
 *      - \c TopBit(0xFF00) returns 16
 */
inline uint32_t TopBit(uint32_t x)
{
    if (x == 0)
        return 0;
    
    uint32_t topBit(0);
    // Find first non-zero
    while (!(x & 0x1))
    {
        x >>= 1;
        ++topBit;
    }
    // Find first non-one
    while (x & 0x1)
    {
        x >>= 1;
        ++topBit;
    }
    return topBit;
}

//-----------------------------------------------------------------------------
/**
 * Returns true if 'x' is power of two (true also for 0 and 1).
 * @note    Found in Steve Baker's "Cute Code" collection.
 */
inline bool IsPowerOfTwo(uint32_t x)
{
    return (x & (x - 1)) == 0;
}

//-----------------------------------------------------------------------------
/**
 * Returns first power of bigger or equal to given integer.
 * @par Example:
 *      - \c NextPowerOfTwo(640)    returns 1024
 *      - \c NextPowerOfTwo(512)    returns 512
 */
inline uint32_t NextPowerOfTwo(uint32_t x)
{
    --x;
    x |= x >> 16;
    x |= x >> 8;
    x |= x >> 4;
    x |= x >> 2;
    x |= x >> 1;
    return x + 1;
}

//-----------------------------------------------------------------------------
/// Returns log2 of given number. Log2(1) = 0, Log2(2) = 1, Log2(4) = 2, etc.
inline uint32_t Log2(uint32_t x)
{
    uint32_t log2(0);
    x >>= 1;
    while (x != 0x0)
    {
        x >>= 1;
        ++log2;
    }
    return log2;
}
    
} // namespace Integer

}

//-----------------------------------------------------------------------------
#endif // #ifndef CORE_INTEGER_H
