#ifndef CORE_RDEDEBUG_H
#define CORE_RDEDEBUG_H

#include "System.h"

namespace rdecore
{

namespace assertion
{
	// @return true if handled/false if not (will halt!)
	typedef bool (*Handler)(const char* expr, const char* file, int line);
	void SetHandler(Handler newHandler);

    /** 
	* @internal Called when assertion fails.
     */
    bool Failure(const char* expr, const char* file, int line);
} // namespace internal

#define RDE_HALT()	do { __debugbreak(); } while (0) 

/**
 * Assertion that works in both debug and release builds.
 * It tests if expr==0 and reports error in such case (exact behavior
 * may be customized, it can throw exception, break to the debugger etc).
 */
#define RDE_ASSERT_ALWAYS(expr)									\
    do															\
    {															\
        if (!(expr))											\
        {														\
            if (!rdecore::assertion::Failure(					\
					"Assertion failed: "						\
					#expr, __FILE__, __LINE__))					\
			{													\
				RDE_HALT();										\
			}													\
        }                                                       \
    }                                                           \
    while (false)

#if RDE_DEBUG
    /// Works only in debug build. @see RDE_ASSERT_ALWAYS
#   define RDE_ASSERT(expr)                     RDE_ASSERT_ALWAYS(expr)
    /**
     * In debug build RDE_ASSUME is the same as RDE_ASSERT, in release
     * it puts the expression in 'whatIfFails' when the test fails.
     * @par Example: 
     * @code
     *  // Now, if ptr == 0 then in debug build we get an assertion,
     *  // and in release we exit function, avoiding the null pointer
     *  // exception.
     *  RDE_ASSUME(ptr != 0, return);
     *  ptr->Foo();
     * @endcode
     */
#   define RDE_ASSUME(expr, whatIfFails)        RDE_ASSERT_ALWAYS(expr)
    /**
     * The only difference between RDE_ASSERT and RDE_VERIFY is that
     * the expression is still evaluated in release mode in RDE_VERIFY.
     * @par Example:
     * @code
     *  // In debug build MyFunction() is called and its result is
     *  // tested, in release mode MyFunction() is not called, because
     *  // the whole expression is eliminated.
     *  RDE_ASSERT(MyFunction());
     *  // In debug MyFunction() is called and its result is
     *  // tested, in release mode MyFunction() is still called, but
     *  // the result is ignored.
     *  RDE_VERIFY(MyFunction());
     * @endcode
     */
#   define RDE_VERIFY(expr)                     RDE_ASSERT_ALWAYS((expr))
#else
#   define RDE_ASSERT(expr)						do { (void)sizeof(expr); } while (0) 
#   define RDE_ASSUME(expr, whatIfFails)        if (!(expr)) {whatIfFails;}
#   define RDE_VERIFY(expr)                     expr
#endif

//-----------------------------------------------------------------------------
/**
 * Compile-time assertion.
 * If expr is zero -> msg will appear in compile-time error message.
 */
//-----------------------------------------------------------------------------
#define RDE_COMPILE_CHECK(expr, msg) \
    typedef char ERROR_##msg[(expr) != 0 ? 1 : -1]


} // rdecore

//-----------------------------------------------------------------------------
#endif // #ifndef CORE_RDEDEBUG_H
