#ifndef CORE_REFCOUNTED_H
#define CORE_REFCOUNTED_H

namespace rdecore
{
//=============================================================================
/**
 * Base/mixin to help with reference counting.
 * Can be used for statically allocated objects (it is not recommened, though),
 * but make sure you DO NOT modify the reference count in such case.
 */
 //=============================================================================
class RefCounted
{
public:
    explicit RefCounted(long initialRefCount = 0);
    /// Copy constructor, initializes reference count to zero.
    RefCounted(const RefCounted&);
    virtual ~RefCounted();

    RefCounted& operator=(const RefCounted&);

    void AddRef();
    /// Decrements reference count by one, calls OnDispose when it reaches 0.
    void Release();
    long GetRefCount() const;

protected:
	virtual void OnDispose() { delete this; }
private:
    long m_refCount;
};

}

namespace boost
{
//@{    
/// Required to make boost::intrusive_ptr work with RefCounted.
//-----------------------------------------------------------------------------
inline void intrusive_ptr_add_ref(rdecore::RefCounted* ptr)
{
    ptr->AddRef();
}
//-----------------------------------------------------------------------------
inline void intrusive_ptr_release(rdecore::RefCounted* ptr)
{
    ptr->Release();
}
//@}
} // namespace boost

//-----------------------------------------------------------------------------
#endif // #ifndef CORE_REFCOUNTED_H
