#include "StackTrace.h"
#include "System.h"

#if (RDE_PLATFORM_WIN)

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <dbghelp.h>
#include <cstdlib>

#pragma comment(lib, "dbghelp.lib")

namespace
{
	struct Internal_ModuleInfo
	{
		DWORD64	baseAddress;
		DWORD64 imageBase;
		DWORD	moduleSize;
		TCHAR	pdbName[_MAX_PATH];
	};
	const int MAX_MODULES	= 256;
	Internal_ModuleInfo	s_loadedModules[MAX_MODULES];
	int					s_numLoadedModules(0);
	HANDLE				s_hSymbolsProcess(0);

	void GetFileFromPath(const char* path, char* file, int fileNameSize)
	{
		char ext[_MAX_EXT] = { 0 };
		_splitpath_s(path, 0, 0, 0, 0, file, fileNameSize, ext, _MAX_EXT);
		strncat_s(file, fileNameSize, ext, _MAX_EXT);
	}

	void ErrorExit(LPTSTR lpszFunction) 
	{ 
		LPVOID lpMsgBuf;
		LPVOID lpDisplayBuf;
		DWORD dw = GetLastError(); 

		FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			dw,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0, NULL );

		lpDisplayBuf = (LPVOID)LocalAlloc(LMEM_ZEROINIT,

			(lstrlen((LPCTSTR)lpMsgBuf)+lstrlen((LPCTSTR)lpszFunction)+40)*sizeof(TCHAR)); 
		wsprintf((LPTSTR)lpDisplayBuf, 
			TEXT("%s failed with error %d: %s"), 
			lpszFunction, dw, lpMsgBuf); 
		MessageBox(NULL, (LPCTSTR)lpDisplayBuf, TEXT("Error"), MB_OK); 

		LocalFree(lpMsgBuf);
		LocalFree(lpDisplayBuf);
		ExitProcess(dw); 
	}

	BOOL CALLBACK EnumerateModule(PCSTR moduleName, DWORD64 moduleBase, ULONG moduleSize,
		PVOID /*userContext*/)
	{
		const HANDLE hCurrentProcess = ::GetCurrentProcess();
		const DWORD64 imageBase = ::SymLoadModule64(hCurrentProcess, NULL, moduleName, NULL,
			moduleBase, moduleSize);
		IMAGEHLP_MODULE64 moduleInfo;
		memset(&moduleInfo, 0, sizeof(IMAGEHLP_MODULE64));
		moduleInfo.SizeOfStruct = sizeof(IMAGEHLP_MODULE64);
		if (::SymGetModuleInfo64(hCurrentProcess, imageBase, &moduleInfo) &&
			s_numLoadedModules < MAX_MODULES)
		{
			Internal_ModuleInfo internalInfo;
			internalInfo.baseAddress	= moduleBase;
			internalInfo.imageBase		= imageBase;
			internalInfo.moduleSize		= moduleSize;
			strncpy_s(internalInfo.pdbName, moduleInfo.LoadedPdbName, sizeof(internalInfo.pdbName));
			s_loadedModules[s_numLoadedModules++] = internalInfo;
		}
		else
		{
			//ErrorExit("EnumModule");
		}
		//::SymUnloadModule64(hCurrentProcess, imageBase);

		return true;
	}

	void UnloadModules()
	{
		for (int i = 0; i < s_numLoadedModules; ++i)
		{
			if (s_loadedModules[i].imageBase != 0)
			{
				if (::SymUnloadModule64(s_hSymbolsProcess, s_loadedModules[i].baseAddress))
					s_loadedModules[i].imageBase = 0;
			}
		}
	}
	void __cdecl UnloadSymbols()
	{
		if (s_hSymbolsProcess != 0)
		{
			UnloadModules();
			::SymCleanup(s_hSymbolsProcess);
			s_hSymbolsProcess = 0;
		}
	}

	void EnumerateModules()
	{
		if (s_numLoadedModules != 0 || s_hSymbolsProcess != 0)
			return;

		const HANDLE hCurrentProcess = ::GetCurrentProcess();
		// Deferred loads conflicts with filling PDB name.
		::SymSetOptions(SYMOPT_UNDNAME /*| SYMOPT_DEFERRED_LOADS*/ | SYMOPT_LOAD_LINES |
			SYMOPT_FAIL_CRITICAL_ERRORS);
		if (::SymInitialize(hCurrentProcess, 0, false))
		{
			s_hSymbolsProcess = hCurrentProcess;
			char modulePath[_MAX_PATH] = { 0 };
			::GetModuleFileName(::GetModuleHandle(0), modulePath, sizeof(modulePath));

			char moduleFile[_MAX_FNAME] = { 0 };
			GetFileFromPath(modulePath, moduleFile, sizeof(moduleFile));
			::EnumerateLoadedModules64(hCurrentProcess, EnumerateModule, moduleFile);

			::atexit(UnloadSymbols);
		}
	}

	bool OpenPDB(HANDLE hProcess, Internal_ModuleInfo& info)
	{
		DWORD64 hModule = ::SymLoadModule64(hProcess, NULL, info.pdbName, NULL, info.baseAddress, info.moduleSize);
		return hModule != 0;
	}
	void LoadModules()
	{
		for (int i = 0; i < s_numLoadedModules; ++i)
		{
			if (s_loadedModules[i].imageBase == 0)
				OpenPDB(s_hSymbolsProcess, s_loadedModules[i]);
		}
	}

	bool SymbolsInit(HANDLE hProcess)
	{
		if (s_hSymbolsProcess == 0)
		{
			::SymSetOptions(SYMOPT_UNDNAME | SYMOPT_DEFERRED_LOADS | SYMOPT_LOAD_LINES);
			if (::SymInitialize(hProcess, 0, false))
			{
				s_hSymbolsProcess = hProcess;
				::atexit(UnloadSymbols);
			}
		}
		return s_hSymbolsProcess != 0;
	}
}

namespace rdecore
{
//-----------------------------------------------------------------------------
int StackTrace::GetCallStack(address_t* callStack, int maxDepth, 
							  int entriesToSkipAtStart)
{
    unsigned int* ebpReg(0);
	unsigned int espReg(0);
    __asm mov [ebpReg], ebp
	__asm mov [espReg], esp

	int numEntries(0);
	int numFilledEntries(0);
    if(ebpReg != 0)
    {
		STACKFRAME64 stackFrame;

        unsigned int frame = 0;
        address_t address = 0;
        frame = ebpReg[0];
        address = ebpReg[1];

        memset(&stackFrame, 0, sizeof(stackFrame));
        stackFrame.AddrPC.Offset = address;
        stackFrame.AddrPC.Mode = AddrModeFlat;
        stackFrame.AddrFrame.Offset = frame;
        stackFrame.AddrFrame.Mode = AddrModeFlat;
		stackFrame.AddrStack.Offset = espReg;
		stackFrame.AddrStack.Mode = AddrModeFlat;

		const HANDLE hProcess = ::GetCurrentProcess();
		const HANDLE hThread = ::GetCurrentThread();
        while (numEntries < maxDepth)
        {
            if (StackWalk64(
                IMAGE_FILE_MACHINE_I386,
                hProcess,
                hThread,
                &stackFrame,
                NULL,
                NULL,
                SymFunctionTableAccess64,
                SymGetModuleBase64,
                NULL))
            {
				if (numEntries >= entriesToSkipAtStart)
				{
					address = static_cast<address_t>(stackFrame.AddrPC.Offset);
					if (address == NULL)
					{
						callStack[numFilledEntries] = 0;
						break;
					}
					else
					{
						callStack[numFilledEntries] = address;
					}
					++numFilledEntries;
				}
                ++numEntries;
            }
            else
            {
                break;
            }
        }
    }
    return numFilledEntries;
}

//-----------------------------------------------------------------------------
void StackTrace::GetSymbolForAddress(address_t address, char* symbol, 
									 int maxSymbolLen)
{
	symbol[0] = '\0';

	::EnumerateModules();
	ULONG64 buffer[(sizeof(SYMBOL_INFO) + MAX_SYM_NAME*sizeof(TCHAR) +
			sizeof(ULONG64) - 1) / sizeof(ULONG64)] = { 0 };
	IMAGEHLP_SYMBOL64* symbolInfo = reinterpret_cast<IMAGEHLP_SYMBOL64*>(buffer);
	symbolInfo->SizeOfStruct = sizeof(IMAGEHLP_SYMBOL64);
	symbolInfo->MaxNameLength = MAX_SYM_NAME;
	::SymGetSymFromAddr64(s_hSymbolsProcess, address, 0, symbolInfo);
	::strncpy_s(symbol, maxSymbolLen, symbolInfo->Name, MAX_SYM_NAME);
}

//-----------------------------------------------------------------------------
void StackTrace::GetFileLineForAddress(address_t address, char* file, 
									   int maxFileLen, int& line)
{
	line = -1;
	file[0] = '\0';

	::EnumerateModules();
	IMAGEHLP_LINE64 lineInfo;
	memset(&lineInfo, 0, sizeof(IMAGEHLP_LINE64));
	lineInfo.SizeOfStruct = sizeof(IMAGEHLP_LINE64);
	DWORD displacement(0);
	if (::SymGetLineFromAddr64(s_hSymbolsProcess, address, &displacement, &lineInfo))
	{
		strncpy_s(file, maxFileLen, lineInfo.FileName, _MAX_PATH);
		line = lineInfo.LineNumber;
	}
	else
	{
		// error!
	}
}

//-----------------------------------------------------------------------------
void StackTrace::GetMainModuleInfo(char* outPdbName, int maxPdbNameLen,
								   unsigned long& outModuleBase)
{
	::EnumerateModules();
	// NOTE: Assumes main module is the first one here... May not always be true,
	// works for now.
	for (int i = 0; i < s_numLoadedModules; ++i)
	{
		if (s_loadedModules[i].pdbName[0] != '\0')
		{
			strcpy_s(outPdbName, maxPdbNameLen, s_loadedModules[i].pdbName);
			outModuleBase = static_cast<unsigned long>(s_loadedModules[i].baseAddress);
			return;
		}
	}
}

//-----------------------------------------------------------------------------
void StackTrace::UnloadSymbols()
{
	::UnloadSymbols();
}

} // rdecore

#else
#	error "NYI!"
#endif

