#include "core/MemManager.h"
#include "core/RdeDebug.h"
#include "core/StackTrace.h"
#include "sockets/WinSocket.h"
#include <algorithm>
#include <conio.h>
#include <cstdio>
#include <memory.h>
#include <windows.h>
#include <vector>

//-----------------------------------------------------------------------------

namespace
{
const unsigned short kListeningPort = 5002;

struct Packet
{
	enum Command
	{
		CMD_PROCESS_ID = 1,
		CMD_MODULE_INFO,
		CMD_ALLOC,
		CMD_FREE,
		CMD_FRAME_START,
		CMD_FRAME_END,
		CMD_TAG_BLOCK,
		CMD_ASSERT,
		CMD_MAX
	};

	unsigned int	command;
	char			data[512];
};

bool SendCommand(Packet::Command cmd, const void* data, int dataSize, WinSocket& socket)
{
	Packet packet;
	packet.command = cmd;
	memcpy(packet.data, data, dataSize);

	//printf("Sending %d\n", cmd);

	return socket.Send(&packet, dataSize + sizeof(packet.command));
}

WinSocket outSocket;
struct AllocInfo
{
	enum { MAX_STACK_DEPTH	= 10 };

	AllocInfo()
	{
		for (int i = 0; i < MAX_STACK_DEPTH; ++i)
			callStack[i] = 0;
	}
	const void*						ptr;
	std::size_t						bytes;
	rdecore::StackTrace::address_t	callStack[MAX_STACK_DEPTH];
};
struct ModuleInfo
{
	unsigned long moduleBase;
	char pdbName[256];
};
struct TagInfo
{
	const void*	ptr;
	char		tag[256];
};
struct AssertInfo
{
	enum { MAX_STACK_DEPTH	= 10 };
	char							expr[256];
	rdecore::StackTrace::address_t	callStack[MAX_STACK_DEPTH];
};

void OnAlloc(const void* ptr, std::size_t bytes, const char*)
{
	if (outSocket.IsOpen())
	{
		AllocInfo info;
		// Skip OnAlloc/Alloc
		const int numEntries = 
			rdecore::StackTrace::GetCallStack(&info.callStack[0], AllocInfo::MAX_STACK_DEPTH, 2);
		std::reverse(&info.callStack[0], &info.callStack[numEntries]);
		/*char symbol[256] = { 0 };
		int line(0);
		for (int i = 0; i < numEntries; ++i)
		{
			printf("%d: %ul\n", i, info.callStack[i]);
			rdecore::StackTrace::GetSymbolForAddress(info.callStack[i], symbol, 256);
			printf("%s\n", symbol);
			rdecore::StackTrace::GetFileLineForAddress(info.callStack[i], symbol, 256, line);
			printf("%s@%d\n", symbol, line);
		}*/

		info.ptr = ptr;
		info.bytes = bytes;
		SendCommand(Packet::CMD_ALLOC, &info, sizeof(info), outSocket);
	}
}
void OnFree(const void* ptr)
{
	if (outSocket.IsOpen())
	{
		AllocInfo info;
		info.ptr = ptr;
		SendCommand(Packet::CMD_FREE, &info, sizeof(info), outSocket);
	}
}

void* Foo()
{
	return new char[3000];
}

void Foobar()
{
	void* block = Foo();
	TagInfo info;
	info.ptr = block;
	strcpy(info.tag, "Test block from foobar");
	SendCommand(Packet::CMD_TAG_BLOCK, &info, sizeof(info), outSocket);
}

bool AssertHandler(const char* expr, const char*, int)
{
	if (outSocket.IsOpen())
	{
		AssertInfo info;
		strcpy_s(info.expr, sizeof(info.expr), expr);
		const int numEntries = 
			rdecore::StackTrace::GetCallStack(&info.callStack[0], AssertInfo::MAX_STACK_DEPTH, 0);
		std::reverse(&info.callStack[0], &info.callStack[numEntries]);
		return SendCommand(Packet::CMD_ASSERT, &info, sizeof(info), outSocket);
	}
	return false;
}

} // namespace

//-----------------------------------------------------------------------------
int __cdecl main()
{
	rdecore::assertion::SetHandler(&AssertHandler);
	rdecore::MemManager::SetOnAllocHandler(&OnAlloc);
	rdecore::MemManager::SetOnFreeHandler(&OnFree);

	ModuleInfo modInfo;
	rdecore::StackTrace::GetMainModuleInfo(modInfo.pdbName, sizeof(modInfo.pdbName), modInfo.moduleBase);

	Foo();

	WinSocket listeningSocket;
	if (!listeningSocket.Create())
		return 1;
	if (!listeningSocket.Bind(kListeningPort))
		return 1;

	printf("Listening...");
	if (!listeningSocket.Listen())
		return 1;


	while (!listeningSocket.WaitForConnection())
	{
		// WAITING!
	}

	if (!listeningSocket.Accept(outSocket))
		return 1;

	printf("Connection made.");
	const DWORD processID = ::GetCurrentProcessId();
	if (!SendCommand(Packet::CMD_PROCESS_ID, &processID, sizeof(processID), outSocket))
		return 1;

	if (!SendCommand(Packet::CMD_MODULE_INFO, &modInfo, sizeof(modInfo), outSocket))
		return 1;

	RDE_ASSERT(!"Just showing off my cool assert system...");

	typedef std::vector<void*>	tBlocks;
	SendCommand(Packet::CMD_FRAME_START, 0, 0, outSocket);
	tBlocks blocks;

	Foo();
	Foobar();
	for (int i = 0; i < 10; ++i)
	{
		char* buf = new char[5000];
		blocks.push_back(buf);
	}
	SendCommand(Packet::CMD_FRAME_END, 0, 0, outSocket);
#if 1
	::srand(132);
	bool done(false);
	while (!done)
	{
		SendCommand(Packet::CMD_FRAME_START, 0, 0, outSocket);
		const int r = rand();

		if ((r & 127) == 0)
		{
			const int numBlocks = rand() % 10;

			for (int i = 0; i < numBlocks; ++i)
			{
				char* buf = new char[r % 35000];
				blocks.push_back(buf);
			}
		}
		else if ((r & 1023) == 0 && !blocks.empty())
		{
			delete[] reinterpret_cast<char*>(blocks.back());
			blocks.pop_back();
		}
		SendCommand(Packet::CMD_FRAME_END, 0, 0, outSocket);
		if (_kbhit() && _getch() == 27)
			done = true;
	}
#endif
	return 0;
}

