#include "WinSocket.h"
#include <cassert>
#include <Winsock2.h>

#pragma comment(lib, "ws2_32.lib")

namespace
{
	SOCKET AsSocket(void* vsocket)
	{
		return reinterpret_cast<SOCKET>(vsocket);
	}
	void* const kInvalidSocket = reinterpret_cast<void*>(INVALID_SOCKET);
}

//-----------------------------------------------------------------------------
int WinSocket::m_refCount(0);

WinSocket::WinSocket()
:	m_socket(kInvalidSocket)
{
	if (++m_refCount == 1)
	{
		WSADATA info;
		::WSAStartup(MAKEWORD(2, 2), &info);
	}
}

//-----------------------------------------------------------------------------
WinSocket::~WinSocket()
{
	Close();

	assert(m_refCount > 0);
	if (--m_refCount <= 0)
	{
		::WSACleanup();
	}
}

//-----------------------------------------------------------------------------
bool WinSocket::Create()
{
	const SOCKET hSocket = ::socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	m_socket = reinterpret_cast<void*>(hSocket);
	return hSocket != INVALID_SOCKET;
}

//-----------------------------------------------------------------------------
bool WinSocket::Connect(const char* hostName, unsigned short port)
{
	hostent* ent = ::gethostbyname(hostName);
	unsigned long addr(0);
	SOCKADDR_IN addrIn;
	if (ent == 0)
	{
		addr = ::inet_addr(hostName);
		if (addr == INADDR_NONE)
			return false;
		else
			addrIn.sin_family = AF_INET;
	}
	else
	{
		memcpy(&addr, ent->h_addr_list[0], sizeof(addr));
		addrIn.sin_family = ent->h_addrtype;
	}
	addrIn.sin_addr.s_addr = addr;
	addrIn.sin_port = ::htons(port);
	memset(addrIn.sin_zero, 0, sizeof(addrIn.sin_zero));
	return ::connect(::AsSocket(m_socket), (const sockaddr*)&addrIn, 
		sizeof(SOCKADDR_IN)) != SOCKET_ERROR;
}

//-----------------------------------------------------------------------------
bool WinSocket::Bind(unsigned short localPort)
{
	SOCKADDR_IN addr;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_family = AF_INET;
	addr.sin_port = ::htons(localPort);
	memset(addr.sin_zero, 0, sizeof(addr.sin_zero));
	return (::bind(::AsSocket(m_socket), (const sockaddr*)&addr, 
		sizeof(SOCKADDR_IN)) != SOCKET_ERROR);
}

//-----------------------------------------------------------------------------
bool WinSocket::WaitForConnection()
{
	if (IsOpen())
	{
		fd_set fdsetRead;
		FD_ZERO(&fdsetRead);
#	pragma warning(push)
	// The implementation of FD_SET use a do { ... }while(0) and thus fire this warning
#	pragma warning(disable: 4127) // conditional expression is consta
		FD_SET(::AsSocket(m_socket), &fdsetRead); 
#	pragma warning(pop)

		if (::select(0, &fdsetRead, 0, 0, 0) > 0)
		{
			if (FD_ISSET(::AsSocket(m_socket), &fdsetRead))
			{
				return true;
			}
		}
	}
	return false;
}

//-----------------------------------------------------------------------------
bool WinSocket::Accept(WinSocket& otherSocket)
{
	SOCKADDR_IN addr;
	int len = sizeof(SOCKADDR_IN);
	memset(&addr, 0, sizeof(SOCKADDR_IN));

	const SOCKET hSocket = ::accept(::AsSocket(m_socket), (sockaddr*)&addr, &len);
	otherSocket.m_socket = reinterpret_cast<void*>(hSocket);
	return hSocket != INVALID_SOCKET;
}

//-----------------------------------------------------------------------------
bool WinSocket::Listen(int numConnections)
{
	return ::listen(::AsSocket(m_socket), numConnections) != SOCKET_ERROR;
}

//-----------------------------------------------------------------------------
bool WinSocket::Send(const void* data, int bytes)
{
	return ::send(::AsSocket(m_socket), reinterpret_cast<const char*>(data), bytes, 0) == bytes;
}

//-----------------------------------------------------------------------------
bool WinSocket::Close()
{
	const bool bSuccess = ::closesocket(::AsSocket(m_socket)) != SOCKET_ERROR;
	m_socket = kInvalidSocket;
	return bSuccess;
}

//-----------------------------------------------------------------------------
bool WinSocket::IsOpen() const
{
	return m_socket != kInvalidSocket;
}
//-----------------------------------------------------------------------------


